
__version__ = '0.6.1'

import click

from .asset.builder import AssetTool
from .tool.cmake import cmake_cli
from .tool.flasher import flash_cli, install_cli, launch_cli
from .tool.metadata import metadata_cli
from .tool.packer import pack_cli
from .tool.relocs import relocs_cli


@click.group()
@click.option('--debug', is_flag=True)
def main(debug):
    pass


for n, c in AssetTool._commands.items():
    main.add_command(c)

main.add_command(cmake_cli)
main.add_command(flash_cli)
main.add_command(install_cli)
main.add_command(launch_cli)
main.add_command(metadata_cli)
main.add_command(pack_cli)
main.add_command(relocs_cli)


@main.command(help='Print version and exit')
def version():
    print(__version__)



from math import ceil

from bitstring import BitArray, Bits, ConstBitStream
from construct import Adapter


class RL:
    @staticmethod
    def repetitions(seq):
        """Input: sequence of values, Output: sequence of (value, repeat count)."""
        i = iter(seq)
        prev = next(i)
        count = 1
        for v in i:
            if v == prev:
                count += 1
            else:
                yield prev, count
                prev = v
                count = 1
        yield prev, count

    @staticmethod
    def compress(data, bit_length):
        """Input: data bytes, bit length, Output: RLE'd bytes"""
        # TODO: This could be made more efficient by encoding the run-length
        # as count-n, ie 0 means a run of n, where n = break_even. Then we
        # can encode longer runs up to 255+n. But this needs changes in the
        # blit engine.
        break_even = ceil(8 / (bit_length + 1))
        bits = BitArray()

        for value, count in RL.repetitions(data):
            while count > break_even:
                chunk = min(count, 0x100)
                bits.append('0b1')
                bits.append(bytes([chunk - 1]))
                count -= chunk
                bits.append(BitArray(uint=value, length=bit_length))
            for x in range(count):
                bits.append('0b0')
                bits.append(BitArray(uint=value, length=bit_length))
        return bits.tobytes()

    @staticmethod
    def decompress(data, bit_length, output_length):
        stream = ConstBitStream(bytes=data)
        result = []
        while len(result) < output_length:
            t = stream.read(1)
            if t:
                count = stream.read(8).uint + 1
            else:
                count = 1
            result.extend([stream.read(bit_length).uint] * count)
        return bytes(result)


class PK:
    @staticmethod
    def compress(data, bit_length):
        return BitArray().join(BitArray(uint=x, length=bit_length) for x in data).tobytes()

    @staticmethod
    def decompress(data, bit_length, num_pixels):
        return bytes(i.uint for i in Bits(bytes=data).cut(bit_length))


packers = {cls.__name__: cls for cls in (PK, RL)}


class ImageCompressor(Adapter):

    def bit_length(self, obj):
        """Compute the required bit length for image data.
        Uses the count of items in the palette to determine how
        densely we can pack the image data.
        """
        if obj.get('type', None) == "RW":
            return 8
        else:
            return max(1, (len(obj['data']['palette']) - 1).bit_length())

    def num_pixels(self, obj):
        return obj['data']['width'] * obj['data']['height']

    def _decode(self, obj, context, path):
        if obj['type'] != 'RW':
            obj['data']['pixels'] = packers[obj['type']].decompress(
                obj['data']['pixels'], self.bit_length(obj), self.num_pixels(obj)
            )
        return obj

    def _encode(self, obj, context, path):
        obj = obj.copy()   # we are going to mutate this, so make a deep copy
        obj['data'] = obj['data'].copy()
        bl = self.bit_length(obj)
        if obj.get('type', None) is None:
            all_comp = [(k, v.compress(obj['data']['pixels'], bl)) for k, v in packers.items()]
            best = min(all_comp, key=lambda x: len(x[1]))
            # Put the best type back into the object.
            obj['type'] = best[0]
            obj['data']['pixels'] = best[1]
        elif obj['type'] != 'RW':
            obj['data']['pixels'] = packers[obj['type']].compress(obj['data']['pixels'], bl)
        return obj



import argparse
import logging
import pathlib
import struct

from PIL import Image


class Colour():
    def __init__(self, colour):
        if type(colour) is Colour:
            self.r, self.g, self.b = colour
        elif len(colour) == 6:
            self.r, self.g, self.b = tuple(bytes.fromhex(colour))
        elif ',' in colour:
            self.r, self.g, self.b = [int(c, 10) for c in colour.split(',')]

    def __getitem__(self, index):
        return [self.r, self.g, self.b][index]

    def __repr__(self):
        return f'Colour{self.r, self.g, self.b}'


class Palette():
    def __init__(self, palette_file=None):
        self.transparent = None
        self.entries = []

        if isinstance(palette_file, Palette):
            self.transparent = palette_file.transparent
            self.entries = palette_file.entries
            return

        if palette_file is not None:
            palette_file = pathlib.Path(palette_file)
            palette_type = palette_file.suffix[1:]

            palette_loader = f'load_{palette_type}'

            fn = getattr(self, palette_loader, None)
            if fn is None:
                self.load_image(palette_file)
            else:
                fn(palette_file, open(palette_file, 'rb').read())

            self.extract_palette()

    def extract_palette(self):
        self.entries = []
        for y in range(self.height):
            for x in range(self.width):
                self.entries.append(self.image.getpixel((x, y)))

    def set_transparent_colour(self, r, g, b):
        if (r, g, b, 0xff) in self.entries:
            self.transparent = self.entries.index((r, g, b, 0xff))
            self.entries[self.transparent] = (r, g, b, 0x00)
            return self.transparent
        return None

    def load_act(self, palette_file, data):
        # Adobe Colour Table .act
        palette = data
        if len(palette) < 772:
            raise ValueError(f'palette {palette_file} is not a valid Adobe .act (length {len(palette)} != 772')

        size, _ = struct.unpack('>HH', palette[-4:])
        self.width = 1
        self.height = size
        self.image = Image.frombytes('RGB', (self.width, self.height), palette).convert('RGBA')

    def load_pal(self, palette_file, data):
        # Pro Motion NG .pal - MS Palette files and raw palette files share .pal suffix
        # Raw files are just 768 bytes
        palette = data
        if len(palette) < 768:
            raise ValueError(f'palette {palette_file} is not a valid Pro Motion NG .pal')
        # There's no length in .pal files, so we just create a 16x16 256 colour palette
        self.width = 16
        self.height = 16
        self.image = Image.frombytes('RGB', (self.width, self.height), palette).convert('RGBA')

    def load_gpl(self, palette_file, data):
        palette = data
        palette = palette.decode('utf-8').strip()
        if not palette.startswith('GIMP Palette'):
            raise ValueError(f'palette {palette_file} is not a valid GIMP .gpl')
        # Split the whole file into palette entries
        palette = palette.split('\r\n')
        # drop 'GIMP Palette' from the first entry
        palette.pop(0)

        # calculate our image width/height here because len(palette)
        # equals the number of palette entries
        self.width = 1
        self.height = len(palette)

        # Split out the palette entries into R, G, B and drop the hex colour
        # This convoluted list comprehension does this while also flatenning to a 1d array
        palette = [int(c) for entry in palette for c in entry.split('\t')[:-1]]
        self.image = Image.frombytes('RGB', (self.width, self.height), bytes(palette)).convert('RGBA')

    def load_image(self, palette_file):
        palette = Image.open(palette_file)
        self.width, self.height = palette.size
        if self.width * self.height > 256:
            raise argparse.ArgumentError(None, f'palette {palette_file} has too many pixels {self.width}x{self.height}={self.width*self.height} (max 256)')
        logging.info(f'Using palette {palette_file} {self.width}x{self.height}')

        self.image = palette.convert('RGBA')

    def quantize_image(self, image, transparent=None, strict=False):
        if strict and len(self) == 0:
            raise TypeError("Attempting to enforce strict colours with an empty palette, did you really want to do this?")
        w, h = image.size
        output_image = Image.new('P', (w, h))
        for y in range(h):
            for x in range(w):
                r, g, b, a = image.getpixel((x, y))
                if transparent is not None and (r, g, b) == tuple(transparent):
                    a = 0x00
                index = self.get_entry(r, g, b, a, strict=strict)
                output_image.putpixel((x, y), index)

        return output_image

    def get_entry(self, r, g, b, a, remap_transparent=True, strict=False):
        if (r, g, b, a) in self.entries:
            index = self.entries.index((r, g, b, a))
            return index
            # Noisy print
            # logging.info(f'Re-mapping ({r}, {g}, {b}, {a}) at ({x}x{y}) to ({index})')
        # Anything with 0 alpha that's not in the palette might as well be the transparent colour
        elif a == 0 and self.transparent is not None:
            return self.transparent
        elif not strict:
            if len(self.entries) < 256:
                self.entries.append((r, g, b, a))
                if a == 0:
                    # Set this as the transparent colour - if we had one we'd have returned it already.
                    self.transparent = len(self.entries) - 1
                return len(self.entries) - 1
            else:
                raise TypeError('Out of palette entries')
        else:
            raise TypeError(f'Colour {r}, {g}, {b}, {a} does not exist in palette!')

    def tostruct(self):
        return [dict(zip('rgba', c)) for c in self.entries]

    def __iter__(self):
        return iter(self.entries)

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, index):
        return self.entries[index]



import binascii
import math

from construct import (Adapter, Bytes, Checksum, Const, GreedyBytes, Int8ul,
                       Int16ul, Int32ub, Int32ul, Optional, PaddedString,
                       Prefixed, PrefixedArray, RawCopy, Rebuild, Struct, len_,
                       this)

from .compression import ImageCompressor


class PaletteCountAdapter(Adapter):
    def _decode(self, obj, context, path):
        if obj == 0:
            obj = 256
        return obj

    def _encode(self, obj, context, path):
        if obj == 256:
            obj = 0
        return obj


class ImageSizeAdapter(Adapter):
    """
    Adds the header and type size to the size field.
    The size field itself is already counted.
    """
    def _decode(self, obj, context, path):
        return obj - 8

    def _encode(self, obj, context, path):
        return obj + 8


struct_blit_pixel = Struct(
    'r' / Int8ul,
    'g' / Int8ul,
    'b' / Int8ul,
    'a' / Int8ul
)

struct_blit_image_compressed = Struct(
    'header' / Const(b'SPRITE'),
    'type' / PaddedString(2, 'ASCII'),
    'data' / Prefixed(ImageSizeAdapter(Int32ul), Struct(
        'width' / Int16ul,
        'height' / Int16ul,
        'format' / Const(0x02, Int8ul),
        'palette' / PrefixedArray(PaletteCountAdapter(Int8ul), struct_blit_pixel),
        'pixels' / GreedyBytes,
    ), includelength=True)
)

struct_blit_image = ImageCompressor(struct_blit_image_compressed)

struct_blit_meta = Struct(
    'header' / Const(b'BLITMETA'),
    'data' / Prefixed(Int16ul, Struct(
        'checksum' / Checksum(
            Int32ul,
            lambda data: binascii.crc32(data),
            this._._.bin.data
        ),
        'date' / PaddedString(16, 'ascii'),
        'title' / PaddedString(25, 'ascii'),
        'description' / PaddedString(129, 'ascii'),
        'version' / PaddedString(17, 'ascii'),
        'author' / PaddedString(17, 'ascii'),
        Const(b'BLITTYPE'),
        'category' / PaddedString(17, 'ascii'),
        'url' / PaddedString(129, 'ascii'),
        'filetypes' / PrefixedArray(Int8ul, PaddedString(5, 'ascii')),
        'icon' / struct_blit_image,
        'splash' / struct_blit_image
    ))
)

struct_blit_meta_standalone = Struct(
    'header' / Const(b'BLITMETA'),
    'data' / Prefixed(Int16ul, Struct(
        'checksum' / Int32ul,
        'date' / PaddedString(16, 'ascii'),
        'title' / PaddedString(25, 'ascii'),
        'description' / PaddedString(129, 'ascii'),
        'version' / PaddedString(17, 'ascii'),
        'author' / PaddedString(17, 'ascii'),
        Const(b'BLITTYPE'),
        'category' / PaddedString(17, 'ascii'),
        'url' / PaddedString(129, 'ascii'),
        'filetypes' / PrefixedArray(Int8ul, PaddedString(5, 'ascii')),
        'icon' / struct_blit_image,
        'splash' / struct_blit_image
    ))
)

struct_blit_bin = Struct(
    'header' / Const(b'BLIT'),
    'render' / Int32ul,
    'update' / Int32ul,
    'init' / Int32ul,
    'length' / Int32ul,
    # The length above is actually the _flash_end symbol from startup_user.s
    # it includes the offset into 0x90000000 (external flash)
    # we mask out the highest nibble to correct this into the actual bin length
    # plus subtract 20 bytes for header, symbol and length dwords
    'bin' / Bytes((this.length & 0x0FFFFFFF) - 20)
)

struct_blit_relo = Struct(
    'header' / Const(b'RELO'),
    'relocs' / PrefixedArray(Int32ul, Struct(
        'reloc' / Int32ul
    ))
)

blit_game = Struct(
    'relo' / Optional(struct_blit_relo),
    'bin' / RawCopy(struct_blit_bin),
    'meta' / Optional(struct_blit_meta)
)

blit_game_with_meta = Struct(
    'relo' / Optional(struct_blit_relo),
    'bin' / RawCopy(struct_blit_bin),
    'meta' / struct_blit_meta
)

blit_game_with_meta_and_relo = Struct(
    'relo' / struct_blit_relo,
    'bin' / RawCopy(struct_blit_bin),
    'meta' / struct_blit_meta
)


def compute_icns_data_length(ctx):
    """Compute the required data length for palette based images.
    We need this computation here so we can use `math.ceil` and
    byte-align the result.
    """
    return math.ceil((ctx.width * ctx.height * ctx.bit_length) / 8)


blit_icns = Struct(
    'header' / Const(b'icns'),
    'size' / Rebuild(Int32ub, len_(this.data) + 16),
    'type' / Const(b'ic07'),  # 128×128 icon in PNG format
    'data_length' / Rebuild(Int32ub, len_(this.data) + 8),
    'data' / Bytes(this.data_length - 8)
)



import logging
import pathlib

import yaml


class YamlLoader():

    def setup_for_config(self, config_file, output_path, files=None):
        self.working_path = pathlib.Path('.')
        if config_file is not None:
            if config_file.is_file():
                self.working_path = config_file.parent
            else:
                logging.warning(f'Unable to find config at {config_file}')
            self.config = yaml.safe_load(config_file.read_text())
            logging.info(f'Using config at {config_file}')

        elif files is not None and output_path is not None:
            self.config = {output_path: {file: {} for file in files}}

        if output_path is not None:
            self.destination_path = output_path
        else:
            self.destination_path = self.working_path



import logging
import pathlib

import click

from ..asset.builder import AssetBuilder, make_symbol_name
from ..asset.writer import AssetWriter
from ..core.yamlloader import YamlLoader


class Packer(YamlLoader):

    def run(self, config, output, files, force):
        if config is None and not files:
            raise click.UsageError('You must supply a config or list of input files.')
        self.targets = []
        self.setup_for_config(config, output, files)

        # Top level of our config is filegroups and general settings
        for target, options in self.config.items():
            output_file = self.working_path / target
            logging.info(f'Preparing output target {output_file}')

            asset_sources = []
            target_options = {}

            for key, value in options.items():
                if key in ('prefix', 'type'):
                    target_options[key] = value

            # Strip high-level options from the dict
            # Leaving just file source globs
            for key in target_options:
                options.pop(key)

            for file_glob, file_options in options.items():
                # Treat the input string as a glob, and get an input filelist
                if type(file_glob) is str:
                    input_files = list(self.working_path.glob(file_glob))
                else:
                    input_files = [file_glob]
                if len(input_files) == 0:
                    logging.warning(f'Input file(s) not found {self.working_path / file_glob}')
                    continue

                # Rewrite a single string option to `name: option`
                # This handles: `inputfile.bin: filename` entries
                if type(file_options) is str:
                    file_options = {'name': file_options}

                elif file_options is None:
                    file_options = {}

                # Handle both an array of options dicts or a single dict
                if type(file_options) is not list:
                    file_options = [file_options]

                for file_opts in file_options:
                    asset_sources.append((input_files, file_opts))

            self.targets.append((
                output_file,
                asset_sources,
                target_options
            ))

        self.destination_path.mkdir(parents=True, exist_ok=True)

        for path, sources, options in self.targets:
            aw = AssetWriter()
            for input_files, file_opts in sources:
                for asset in self.build_assets(input_files, self.working_path, prefix=options.get('prefix'), **file_opts):
                    aw.add_asset(*asset)

            aw.write(options.get('type'), self.destination_path / path.name, force=force)

    def build_assets(self, input_files, working_path, name=None, type=None, prefix=None, **builder_options):
        if type is None:
            # Glob files all have the same suffix, so we only care about the first one
            try:
                typestr = AssetBuilder.guess_builder(input_files[0])
            except TypeError:
                logging.warning(f'Unable to guess type, assuming raw/binary {input_files[0]}.')
                typestr = 'raw/binary'
        else:
            typestr = type

        input_type, input_subtype = typestr.split('/')
        builder = AssetBuilder._by_name[input_type]

        # Now we know our target builder, one last iteration through the options
        # allows some pre-processing stages to remap paths or other idiosyncrasies
        # of the yml config format.
        # Currently the only option we need to do this on is 'palette' for images.
        for option in ['palette']:
            try:
                if not pathlib.Path(builder_options[option]).is_absolute():
                    builder_options[option] = working_path / builder_options[option]
            except KeyError:
                pass

        for file in input_files:
            symbol_name = make_symbol_name(
                base=name, working_path=working_path, input_file=file,
                input_type=input_type, input_subtype=input_subtype, prefix=prefix
            )

            yield symbol_name, builder.from_file(file, input_subtype, **builder_options)
            logging.info(f' - {typestr} {file} -> {symbol_name}')


@click.command('pack', help='Pack a collection of assets for 32Blit')
@click.option('--config', type=pathlib.Path, help='Asset config file')
@click.option('--output', type=pathlib.Path, help='Name for output file(s) or root path when using --config')
@click.option('--files', multiple=True, type=pathlib.Path, help='Input files')
@click.option('--force', is_flag='store_true', help='Force file overwrite')
def pack_cli(config, output, files, force):
    Packer().run(config, output, files, force)



import io
import struct

import click
import freetype
from PIL import Image

from ..builder import AssetBuilder, AssetTool

font_typemap = {
    'image': {
        '.png': False,
        '.gif': False,
    },
    'font': {
        # possibly other freetype supported formats...
        '.ttf': True,
    }
}


def process_image_font(data, num_chars, height, horizontal_spacing, space_width):
    # Since we already have bytes, we need to pass PIL an io.BytesIO object
    image = Image.open(io.BytesIO(data)).convert('1')
    w, h = image.size

    if height != 0 and height != h:
        raise TypeError("Specified height does not match image height")

    char_width = w // num_chars
    char_height = h

    font_data = []
    font_w = []  # per character width for variable-width mode

    for c in range(0, num_chars):
        char_w = 0

        for x in range(0, char_width):
            byte = 0

            for y in range(0, h):
                bit = y % 8

                # next byte
                if bit == 0 and y > 0:
                    font_data.append(byte)
                    byte = 0

                if image.getpixel((x + c * char_width, y)) != 0:
                    byte |= 1 << bit
                    if x + 1 > char_w:
                        char_w = x + 1

            font_data.append(byte)

        if c == 0:  # space
            font_w.append(space_width)
        else:
            font_w.append(char_w + horizontal_spacing)

    return font_data, font_w, char_width, char_height


def process_ft_font(data, num_chars, base_char, height):
    if height == 0:
        raise TypeError("Height must be specified for font files")

    face = freetype.Face(io.BytesIO(data))

    # request height
    face.set_pixel_sizes(0, height)

    char_width = 0
    char_height = 0
    min_y = height
    font_w = []

    # measure the actual size of the characters (may not match requested)
    for c in range(0, num_chars):
        face.load_char(c + base_char, freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)

        font_w.append(face.glyph.advance.x >> 6)

        if face.glyph.bitmap.width + face.glyph.bitmap_left > char_width:
            char_width = face.glyph.bitmap.width + face.glyph.bitmap_left

        if (height - face.glyph.bitmap_top) + face.glyph.bitmap.rows > char_height:
            char_height = height - face.glyph.bitmap_top + face.glyph.bitmap.rows

        if height - face.glyph.bitmap_top < min_y:
            min_y = height - face.glyph.bitmap_top

    char_height -= min_y  # trim empty space at the top

    font_data = []

    # now do the conversion
    for c in range(0, num_chars):
        face.load_char(c + base_char, freetype.FT_LOAD_RENDER | freetype.FT_LOAD_TARGET_MONO)

        x_off = face.glyph.bitmap_left
        y_off = height - face.glyph.bitmap_top - min_y

        for x in range(0, char_width):
            byte = 0

            for y in range(0, char_height):
                bit = y % 8

                # next byte
                if bit == 0 and y > 0:
                    font_data.append(byte)
                    byte = 0

                if x < x_off or x - x_off >= face.glyph.bitmap.width or y < y_off or y - y_off >= face.glyph.bitmap.rows:
                    continue

                # freetype monochrome bitmaps are the other way around
                if face.glyph.bitmap.buffer[(x - x_off) // 8 + (y - y_off) * face.glyph.bitmap.pitch] & (
                        0x80 >> ((x - x_off) & 7)):
                    byte |= 1 << bit

            font_data.append(byte)

    return font_data, font_w, char_width, char_height


@AssetBuilder(typemap=font_typemap)
def font(data, subtype, num_chars=96, base_char=ord(' '), height=0, horizontal_spacing=1, vertical_spacing=1, space_width=3):
    if subtype == 'image':
        font_data, font_w_data, char_width, char_height = process_image_font(
            data, num_chars, height, horizontal_spacing, space_width
        )
    elif subtype == 'font':
        font_data, font_w_data, char_width, char_height = process_ft_font(
            data, num_chars, base_char, height
        )
    else:
        raise TypeError(f'Unknown subtype {subtype} for font.')

    head_data = struct.pack('<BBBB', num_chars, char_width, char_height, vertical_spacing)

    data = bytes('FONT', encoding='utf-8')
    data += head_data
    data += bytes(font_w_data)
    data += bytes(font_data)

    return data


@AssetTool(font, 'Convert fonts for 32Blit')
@click.option('--height', type=int, default=0, help='Font height (calculated from image if not specified)')
@click.option('--horizontal-spacing', type=int, default=1, help='Additional space between characters for variable-width mode')
@click.option('--vertical-spacing', type=int, default=1, help='Space between lines')
@click.option('--space-width', type=int, default=3, help='Width of the space character')
def font_cli(input_file, input_type, **kwargs):
    return font.from_file(input_file, input_type, **kwargs)



from ..formatter import AssetFormatter


@AssetFormatter(extensions=('.raw', '.bin'))
def raw_binary(symbol, data):
    return {None: data}


@raw_binary.joiner
def raw_binary(path, fragments):
    return {None: b''.join(fragments[None])}



import struct
import logging

import click

from ..builder import AssetBuilder, AssetTool
from .raw import csv_to_list

map_typemap = {
    'tiled': {
        '.tmx': True,
        '.raw': False,
    },
}


def tiled_to_binary(data, empty_tile, output_struct):
    from xml.etree import ElementTree as ET
    root = ET.fromstring(data)
    layers = root.findall('layer')
    layer_data = []
    # Sort layers by ID (since .tmx files can have them in arbitrary orders)
    layers.sort(key=lambda l: int(l.get('id')))

    use_16bits = False

    for layer_csv in layers:
        layer = csv_to_list(layer_csv.find('data').text, 10)
        # Shift 1-indexed tiles to 0-indexed, and remap empty tile (0) to specified index
        layer = [empty_tile if i == 0 else i - 1 for i in layer]

        if max(layer) > 255 and not use_16bits:
            # Let's assume it's got 2-byte tile indices
            logging.info('Found a tile index > 255, using 16bit tile sizes!')
            use_16bits = True

        # Always build up a 1d array of layer data
        layer_data += layer

    if use_16bits:
        layer_data = struct.pack(f'<{len(layer_data)}H', *layer_data)
    else:
        layer_data = struct.pack(f'<{len(layer_data)}B', *layer_data)

    if output_struct:  # Fancy struct
        layer_count = len(layers)
        width = int(root.get("width"))
        height = int(root.get("height"))

        return struct.pack(
            '<4sBHHH',
            bytes('MTMX', encoding='utf-8'),
            empty_tile,
            width,
            height,
            layer_count
        ) + layer_data

    else:
        # Just return the raw layer data
        return layer_data


@AssetBuilder(typemap=map_typemap)
def map(data, subtype, empty_tile=0, output_struct=False):
    if subtype == 'tiled':
        return tiled_to_binary(data, empty_tile, output_struct)


@AssetTool(map, 'Convert popular tilemap formats for 32Blit')
@click.option('--empty-tile', type=int, default=0, help='Remap .tmx empty tiles')
@click.option('--output-struct', type=bool, default=False, help='Output .tmx as struct with level width/height, etc')
def map_cli(input_file, input_type, **kwargs):
    return map.from_file(input_file, input_type, **kwargs)



import pathlib
import struct

import click
from elftools.elf.elffile import ELFFile
from elftools.elf.enums import ENUM_RELOC_TYPE_ARM


@click.command('relocs', help='Prepend relocations to a game binary')
@click.option('--bin-file', type=pathlib.Path, help='Input .bin', required=True)
@click.option('--elf-file', type=pathlib.Path, help='Input .elf', required=True)
@click.option('--output', type=pathlib.Path, help='Output file', required=True)
def relocs_cli(bin_file, elf_file, output):
    with open(elf_file, 'rb') as f:
        elffile = ELFFile(f)

        # get addresses of GOT values that need patched
        got_offsets = get_flash_addr_offsets(elffile.get_section_by_name('.got'))

        # and the init/fini arrays
        init_offsets = get_flash_addr_offsets(elffile.get_section_by_name('.init_array'))
        fini_offsets = get_flash_addr_offsets(elffile.get_section_by_name('.fini_array'))

        reloc_offsets = []

        # find sidata/sdata
        sidata = 0
        sdata = 0
        relocs = elffile.get_section_by_name('.rel.text')
        symtable = elffile.get_section(relocs['sh_link'])
        for reloc in relocs.iter_relocations():
            symbol = symtable.get_symbol(reloc['r_info_sym'])
            if symbol.name == '_sidata':
                sidata = symbol['st_value']
            elif symbol.name == '_sdata':
                sdata = symbol['st_value']

            if sidata and sdata:
                break

        assert(sidata != 0 and sdata != 0)

        # get all .data relocations
        relocs = elffile.get_section_by_name('.rel.data')
        symtable = elffile.get_section(relocs['sh_link'])

        for reloc in relocs.iter_relocations():
            symbol = symtable.get_symbol(reloc['r_info_sym'])

            if reloc['r_info_type'] != ENUM_RELOC_TYPE_ARM['R_ARM_ABS32']:
                continue

            # doesn't point to flash
            if symbol['st_value'] < 0x90000000:
                continue

            # map RAM address back to flash
            flash_offset = (reloc['r_offset'] - sdata) + sidata

            assert((flash_offset & 3) == 0)

            reloc_offsets.append(flash_offset)

        with open(output, 'wb') as out_f:
            all_offsets = got_offsets + init_offsets + fini_offsets + reloc_offsets
            all_offsets.sort()

            out_f.write(b"RELO")
            out_f.write(struct.pack("<L", len(all_offsets)))
            for off in all_offsets:
                out_f.write(struct.pack("<L", off))

            out_f.write(open(bin_file, 'rb').read())


def get_flash_addr_offsets(section):
    section_data = section.data()
    section_addrs = struct.unpack(f'<{len(section_data) // 4}I', section_data)
    section_offsets = []

    for i, addr in enumerate(section_addrs):
        # filter out non-flash
        if addr < 0x90000000:
            continue

        # offset to this address in the section
        section_offsets.append(section['sh_addr'] + i * 4)

    return section_offsets



import functools
import logging
import pathlib
import struct
import time

import serial.tools.list_ports
from construct.core import ConstructError
from tqdm import tqdm

from ..core.struct import struct_blit_meta_standalone


class BlitSerialException(Exception):
    pass


class BlitSerialWrongModeException(BlitSerialException):
    pass


def firmware_command(fn):
    """Check if we are in the firmware before running commands that require it."""
    @functools.wraps(fn)
    def _decorated(self, *args, **kwargs):
        # if self.status != 'firmware':
        #    # Note: we could instead reset_to_firmware here.
        #    raise BlitSerialWrongModeException("32Blit is not in firmware mode. Please exit game or reset.")
        return fn(self, *args, **kwargs)

    return _decorated


class BlitSerial(serial.Serial):
    def __init__(self, port):
        super().__init__(port, timeout=5)

    @classmethod
    def find_comport(cls):
        ret = []
        for comport in serial.tools.list_ports.comports():
            if comport.vid == 0x0483 and comport.pid == 0x5740:
                logging.info(f'Found 32Blit on {comport.device}')
                ret.append(comport.device)

        if ret:
            return ret

        raise RuntimeError('Unable to find 32Blit')

    @classmethod
    def validate_comport(cls, device):
        if device.lower() == 'auto':
            return cls.find_comport()[:1]
        if device.lower() == 'all':
            return cls.find_comport()

        for comport in serial.tools.list_ports.comports():
            if comport.device == device:
                if comport.vid == 0x0483 and comport.pid == 0x5740:
                    logging.info(f'Found 32Blit on {comport.device}')
                    return [device]
                # if the user asked for a port with no vid/pid assume they know what they're doing
                elif comport.vid is None and comport.pid is None:
                    logging.info(f'Using unidentified port {comport.device}')
                    return [device]
        raise RuntimeError(f'Unable to find 32Blit on {device}')

    @property
    def status(self):
        self.write(b'32BLINFO\x00')
        response = self.read(8)
        if response == b'':
            raise RuntimeError('Timeout waiting for 32Blit status.')
        return 'game' if response == b'32BL_EXT' else 'firmware'

    def reset(self, timeout=5.0):
        self.write(b'32BL_RST\x00')
        self.flush()
        self.close()
        time.sleep(1.0)
        t_start = time.time()
        while time.time() - t_start < timeout:
            try:
                self.open()
                return
            except BlitSerialException:
                time.sleep(0.1)
        raise RuntimeError(f"Failed to connect to 32Blit on {serial.port} after reset")

    def reset_to_firmware(self, timeout=5.0):
        if self.status != 'firmware':
            self.reset(timeout)

    @firmware_command
    def send_file(self, file, drive, directory=pathlib.PurePosixPath('/'), auto_launch=False):
        sent_byte_count = 0
        chunk_size = 64
        file_name = file.name
        file_size = file.stat().st_size

        if drive == 'sd':
            logging.info(f'Saving {file} ({file_size} bytes) as {file_name} in {directory}')
            command = f'32BLSAVE{directory}/{file_name}\x00{file_size}\x00'
        elif drive == 'flash':
            logging.info(f'Flashing {file} ({file_size} bytes)')
            command = f'32BLPROG{file_name}\x00{file_size}\x00'
        else:
            raise TypeError(f'Unknown destination {drive}.')

        self.reset_output_buffer()
        self.write(command.encode('ascii'))

        with open(file, 'rb') as file:
            progress = tqdm(total=file_size, desc="Flashing...", unit_scale=True, unit_divisor=1024, unit="B", ncols=70, dynamic_ncols=True)
            while sent_byte_count < file_size:
                # if we received something, there was an error
                if self.in_waiting:
                    progress.close()
                    break

                data = file.read(chunk_size)
                self.write(data)
                self.flush()
                sent_byte_count += chunk_size
                progress.update(chunk_size)

        response = self.read(8)

        if response == b'32BL__OK':
            # get block for flash
            if drive == 'flash':
                block, = struct.unpack('<H', self.read(2))

                # do launch if requested
                if auto_launch:
                    self.launch(f'flash:/{block}')
            elif auto_launch:
                self.launch(f'{directory}/{file_name}')

            logging.info(f'Wrote {file_name} successfully')
        else:
            response += self.read(self.in_waiting)
            raise RuntimeError(f"Failed to save/flash {file_name}: {response.decode()}")

    @firmware_command
    def erase(self, offset):
        self.write(b'32BLERSE\x00')
        self.write(struct.pack("<I", offset))

    @firmware_command
    def list(self):
        self.write(b'32BL__LS\x00')
        offset_str = self.read(4)

        while offset_str != '' and offset_str != b'\xff\xff\xff\xff':
            offset, = struct.unpack('<I', offset_str)

            size, = struct.unpack('<I', self.read(4))
            meta_head = self.read(10)
            meta_size, = struct.unpack('<H', meta_head[8:])

            meta = None
            if meta_size:
                size += meta_size + 10
                try:
                    meta = struct_blit_meta_standalone.parse(meta_head + self.read(meta_size))
                except ConstructError:
                    pass

            yield (meta, offset, size)

            offset_str = self.read(4)

    @firmware_command
    def launch(self, filename):
        self.write(f'32BLLNCH{filename}\x00'.encode('ascii'))


import binascii
import io
import logging
import pathlib
import struct
from datetime import datetime

import click
from construct.core import StreamError
from PIL import Image

from ..asset.builder import AssetBuilder
from ..core.struct import (blit_game, blit_game_with_meta,
                           blit_game_with_meta_and_relo, blit_icns,
                           struct_blit_image, struct_blit_pixel)
from ..core.yamlloader import YamlLoader


class Metadata(YamlLoader):

    def prepare_image_asset(self, name, config, working_path):
        image_file = pathlib.Path(config.get('file', ''))
        if not image_file.is_file():
            image_file = working_path / image_file
        if not image_file.is_file():
            raise ValueError(f'{name} "{image_file}" does not exist!')
        return AssetBuilder._by_name['image'].from_file(image_file, None)

    def blit_image_to_pil(self, image):
        data = b''.join(struct_blit_pixel.build(image.data.palette[i]) for i in image.data.pixels)
        return Image.frombytes("RGBA", (image.data.width, image.data.height), data)

    def binary_size(self, bin):
        return struct.unpack('<I', bin[16:20])[0] & 0xffffff

    def checksum(self, bin):
        return struct.pack('<I', binascii.crc32(bin))

    def build_icns(self, config, working_path):
        image_file = pathlib.Path(config.get('file', ''))
        if not image_file.is_file():
            image_file = working_path / image_file
        if not image_file.is_file():
            raise ValueError(f'splash "{image_file}" does not exist!')

        blit_icon = Image.frombytes("P", (22, 22), b'\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x00\x00\x00\x00\x01\x01\x01\x01\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x00\x00\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01\x02\x02\x01\x01\x01\x01\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x00\x00\x01\x01\x01\x01')
        blit_icon.putpalette([0, 0, 0, 255, 255, 255, 0, 255, 0])
        blit_icon = blit_icon.convert("RGBA")

        source = Image.open(image_file).convert("RGBA")

        image = Image.new("RGB", (128, 128), (0, 0, 0))

        image.paste(source, (0, 0))
        image.paste(blit_icon, (101, 101), blit_icon)

        image_bytes = io.BytesIO()
        image.convert("RGB").save(image_bytes, format="PNG")
        image_bytes.seek(0)

        del image

        return blit_icns.build({'data': image_bytes.read()})

    def run(self, config, icns, file, force, dump_images):
        if file is None and config is None:
            raise click.UsageError('the following arguments are required: --config and/or --file')

        if file and not file.is_file():
            raise ValueError(f'Unable to find bin file at {file}')

        icon = b''
        splash = b''

        game = None

        if file:
            bin = open(file, 'rb').read()
            try:
                game = blit_game.parse(bin)
            except StreamError:
                raise ValueError(f'Invalid 32blit binary file {file}')

        # No config supplied, so dump the game info
        if config is None:
            print(f'\nParsed:      {file.name} ({game.bin.length:,} bytes)')
            if game.relo is not None:
                print(f'Relocations: Yes ({len(game.relo.relocs)})')
            else:
                print('Relocations: No')
            if game.meta is not None:
                print('Metadata:    Yes')
                for field in ['title', 'description', 'version', 'author', 'category', 'url']:
                    print(f'{field.title()+":":13s}{getattr(game.meta.data, field)}')
                if len(game.meta.data.filetypes) > 0:
                    print('    Filetypes:   ')
                    for filetype in game.meta.data.filetypes:
                        print('       ', filetype)
                if game.meta.data.icon is not None:
                    game_icon = game.meta.data.icon
                    print(f'    Icon:        {game_icon.data.width}x{game_icon.data.height} ({len(game_icon.data.palette)} colours) ({game_icon.type})')
                    if dump_images:
                        image_icon = self.blit_image_to_pil(game_icon)
                        image_icon_filename = file.with_suffix(".icon.png")
                        image_icon.save(image_icon_filename)
                        print(f'    Dumped to:   {image_icon_filename}')
                if game.meta.data.splash is not None:
                    game_splash = game.meta.data.splash
                    print(f'    Splash:      {game_splash.data.width}x{game_splash.data.height} ({len(game_splash.data.palette)} colours) ({game_splash.type})')
                    if dump_images:
                        image_splash = self.blit_image_to_pil(game_splash)
                        image_splash_filename = file.with_suffix('.splash.png')
                        image_splash.save(image_splash_filename)
                        print(f'    Dumped to:   {image_splash_filename}')
            else:
                print('Metadata:    No')
            print('')
            return

        self.setup_for_config(config, None)

        if 'icon' in self.config:
            icon = self.prepare_image_asset('icon', self.config['icon'], self.working_path)
        else:
            raise ValueError('An 8x8 pixel icon is required!"')

        if 'splash' in self.config:
            splash = self.prepare_image_asset('splash', self.config['splash'], self.working_path)
            if icns is not None:
                if not icns.is_file() or force:
                    open(icns, 'wb').write(self.build_icns(self.config['splash'], self.working_path))
                    logging.info(f'Saved macOS icon to {icns}')
        else:
            raise ValueError('A 128x96 pixel splash is required!"')

        if not game:
            return

        title = self.config.get('title', None)
        description = self.config.get('description', '')
        version = self.config.get('version', None)
        author = self.config.get('author', None)
        url = self.config.get('url', '')
        category = self.config.get('category', 'none')
        filetypes = self.config.get('filetypes', [])
        if type(filetypes) is str:
            filetypes = filetypes.split(' ')

        if title is None:
            raise ValueError('Title is required!')

        if version is None:
            raise ValueError('Version is required!')

        if author is None:
            raise ValueError('Author is required!')

        if len(title) > 24:
            raise ValueError('Title should be a maximum of 24 characters!"')

        if len(description) > 128:
            raise ValueError('Description should be a maximum of 128 characters!')

        if len(version) > 16:
            raise ValueError('Version should be a maximum of 16 characters! eg: "v1.0.2"')

        if len(author) > 16:
            raise ValueError('Author should be a maximum of 16 characters!')

        if len(category) > 16:
            raise ValueError('Category should be a maximum of 16 characters!')

        if len(url) > 128:
            raise ValueError('URL should be a maximum of 128 characters!')

        if len(filetypes) > 0:
            for filetype in filetypes:
                if len(filetype) > 4:
                    raise ValueError('Filetype should be a maximum of 4 characters! (Hint, don\'t include the .)')

        if game.meta is not None:
            if not force:
                logging.critical(f'Refusing to overwrite metadata in {file}')
                return 1

        game.meta = {
            'data': {
                'date': datetime.now().strftime("%Y%m%dT%H%M%S"),
                'title': title,
                'description': description,
                'version': version,
                'author': author,
                'category': category,
                'filetypes': filetypes,
                'url': url,
                'icon': struct_blit_image.parse(icon),
                'splash': struct_blit_image.parse(splash)
            }
        }

        # Force through a non-optional builder if relo symbols exist
        # since we might have modified them and want to hit parser errors
        # if something has messed up
        if game.relo is not None:
            bin = blit_game_with_meta_and_relo.build(game)
        else:
            bin = blit_game_with_meta.build(game)

        logging.info(f'Adding metadata to {file}')
        open(file, 'wb').write(bin)

        return 0


@click.command('metadata', help='Tag a 32Blit .blit file with metadata')
@click.option('--config', type=pathlib.Path, help='Metadata config file')
@click.option('--icns', type=pathlib.Path, help='macOS icon output file')
@click.option('--file', type=pathlib.Path, help='Input file')
@click.option('--force', is_flag=True, help='Force file overwrite')
@click.option('--dump-images', is_flag=True, help='Dump images from metadata')
def metadata_cli(config, icns, file, force, dump_images):
    Metadata().run(config, icns, file, force, dump_images)



import importlib
import pkgutil

from . import formatters


class AssetFormatter():
    _by_name = {}
    _by_extension = {}

    def __init__(self, components=None, extensions=None):
        self.components = components if components else (None, )
        self.extensions = extensions

    def __call__(self, fragment_func):
        """Decorator method to create a formatter instance from a fragment function."""
        self.name = fragment_func.__name__
        self.fragments = fragment_func
        return self

    def joiner(self, join_func):
        """Decorator method to attach a join function to a formatter and register it."""
        self.join = join_func
        # Now we have all we need to register.
        self._by_name[self.name] = self
        for ext in self.extensions:
            self._by_extension[ext] = self
        return self

    def __repr__(self):
        return self.name

    @staticmethod
    def fragments(symbol, data):
        raise NotImplementedError

    @staticmethod
    def join(ext, filename, data):
        raise NotImplementedError

    @classmethod
    def names(cls):
        return cls._by_name.keys()

    @classmethod
    def parse(cls, value):
        """Fetch a formatter by name."""
        if isinstance(value, cls):
            return value
        else:
            try:
                return cls._by_name[value]
            except KeyError:
                raise ValueError(f'Invalid format {value}, choices {cls.names()}.')

    @classmethod
    def guess(cls, path):
        """Fetch a formatter which can generate the requested filename."""
        try:
            return cls._by_extension[path.suffix]
        except KeyError:
            raise TypeError(f"Unable to identify format for {path}.")


# Load all the implementations dynamically.
for loader, module_name, is_pkg in pkgutil.walk_packages(formatters.__path__, formatters.__name__ + '.'):
    # We don't need to import anything from the modules. We just need to load them.
    # This will cause the decorators to run, which registers the formatters.
    importlib.import_module(module_name, formatters.__name__)
